# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the path splitting processor."""

import json
import os
import unittest
from unittest.mock import patch

from grove.models import ProcessorConfig
from grove.processors import split_path
from tests import mocks


class ProcessorPathMapperTestCase(unittest.TestCase):
    """Implements unit tests for the path splitting processor."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Setup the processor and associated configuration for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))

        # Create a mapping compatible with the Okta system log fixture.
        self.processor = split_path.Handler(
            ProcessorConfig(
                name="Fan Out",
                processor="split_path",
                source="events",
            )
        )

    def test_split_path(self):
        """Ensure path splitting operates as expected."""
        # Load and process the fixture.
        entries = json.load(
            open(
                os.path.join(self.dir, "fixtures/gsuite/activities/003.json"),
                "r",
            )
        )

        # Strip the outer items list from the input logs, as Grove will have stripped
        # this already.
        entries = entries.get("items", [])

        # Confirm that the initial log entry has one entry.
        self.assertEqual(len(entries), 1)

        # Process the entries.
        records = []
        for entry in entries:
            records.extend(self.processor.process(entry))

        # Confirm that five records resulted after splitting.
        self.assertEqual(len(records), 5)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove processor to filter (delete) entire log entries based JMESPath queries.

This processor is intended to allow dropping of log entries when a set of criteria is
met. This may be used to assist in reducing outputting noisy log entries where a given
vendor does not provide a mechanism for filtering events.
"""

import jmespath
from typing import Any, Dict, List

from pydantic import Extra

from grove.models import ProcessorConfig
from grove.processors import BaseProcessor


class Handler(BaseProcessor):
    """Filter (delete) log entries based on JMESPath queries.

    If any of the configured filters match a given log entry (return True), then the log
    entry will be dropped. Queries are evaluated against log entries in the order that
    they are defined, and the 'first match wins'.
    """

    class Configuration(ProcessorConfig, extra=Extra.forbid):
        """Expresses the configuration and associated validators for the processor."""

        # Filters defines a list of JMESPath queries to be evaluated against each log
        # entry in order.
        filters: List[str]

    def process(self, entry: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Drop log entries which do not match all of the configured JMESPath queries.

        :param entry: A collected log entry.

        :return: The processed log entry, or an empty array if the log entry should be
            dropped.
        """
        for filter in self.configuration.filters:
            if jmespath.search(filter, entry):
                return []

        return [entry]



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove processor to filter (delete) fields from log entries based on provided paths.

This processor is intended to allow removal of superfluous or duplicated data from
log entries. This may be used after a processing stage to remove the original source
data, or used to prune down a log entry from a particularly verbose vendor.
"""

from typing import Any, Dict, List

from pydantic import Extra

from grove.helpers import parsing
from grove.models import ProcessorConfig
from grove.processors import BaseProcessor


class Handler(BaseProcessor):
    """Filter (delete) fields from log entries based on provided paths."""

    class Configuration(ProcessorConfig, extra=Extra.forbid):
        """Expresses the configuration and associated validators for the processor."""

        # Source defines a list of paths to field to drop (delete). These should be
        # defined as a JMESPaths.
        sources: List[str]

    def process(self, entry: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Attempt to drop a configured field from the log entry.

        :param entry: A collected log entry.

        :return: The processed log entry, with fields dropped.
        """
        for source in self.configuration.sources:
            entry = parsing.update_path(
                entry,
                parsing.quote_aware_split(source),
                None,
            )

        return [entry]



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the 1Password Event Item Usage collector."""

import os
import re
import unittest
from unittest.mock import patch

import responses

from grove.connectors.onepassword.events_itemusages import Connector
from grove.models import ConnectorConfig
from tests import mocks


class OnePasswordItemUsageEventTestCase(unittest.TestCase):
    """Implements unit tests for the 1Password Item Usage Event collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                identity="examplecorp",
                key="0123456789",
                name="examplecorp",
                connector="test",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @responses.activate
    def test_client_rate_limit(self):
        """Ensure ratelimit waiting is working as expected."""
        # Rate limit the first request.
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=429,
            content_type="application/json",
            body=bytes(),
        )

        # Succeed on the second.
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/onepassword/events_itemusages/001.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # Ensure we sleep appropriately on rate-limit.
        with patch("time.sleep", return_value=None) as mock_sleep:
            self.connector.run()
            mock_sleep.assert_called_with(1)

    @responses.activate
    def test_collect_no_pagination(self):
        """Ensure collection without pagination is working as expected."""
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/onepassword/events_itemusages/001.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # Ensure only a single value is returned, and the pointer is properly set.
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 3)
        self.assertEqual(self.connector.pointer, "2020-06-11T16:42:55-03:00")

    @responses.activate
    def test_collect_pagination(self):
        """Ensure collection with pagination is working as expected."""
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/onepassword/events_itemusages/002.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(
                        self.dir, "fixtures/onepassword/events_itemusages/003.json"
                    ),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # Ensure only a single value is returned, and the pointer is properly set.
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 2)
        self.assertEqual(self.connector.pointer, "2020-06-11T16:52:55-03:00")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""PagerDuty Audit API client."""

import logging
import time
from typing import Dict, Optional

import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://api.pagerduty.com"


class Client:
    def __init__(
        self,
        token: Optional[str] = None,
        retry: Optional[bool] = True,
    ):
        """Setup a new client.

        :param token: The PagerDuty API Key.
        :param retry: Whether to automatically retry if recoverable errors are
            encountered, such as rate-limiting.
        """
        self.retry = retry
        self.logger = logging.getLogger(__name__)
        self.headers = {
            "Accept": "application/vnd.pagerduty+json;version=3",
            "Authorization": f"Token token={token}",
            "Content-Type": "application/json",
        }

    def _get(
        self,
        url: str,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP GET against.
        :param params: HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        while True:
            try:
                response = requests.get(url, headers=self.headers, params=params)
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as err:
                # Retry on rate-limit, but only if requested.
                if getattr(err.response, "status_code", None) == 429:
                    self.logger.warning("Rate-limit was exceeded during request")
                    if self.retry:
                        time.sleep(1)
                        continue
                    else:
                        raise RateLimitException(err)

                raise RequestFailedException(err)

        return HTTPResponse(headers=response.headers, body=response.json())

    def get_records(
        self,
        since: Optional[str] = None,
        cursor: Optional[str] = None,
        limit: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of audit logs which match the provided filters.

        :param since: The ISO8601 format of the most recent event to fetch (inclusive).
        :param cursor: The cursor to use when paging.
        :param limit: Number of audit records per request.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        # See psf/requests issue #2651 for why we can happily pass in None values
        # and not have the request key added to the URI.
        result = self._get(
            f"{API_BASE_URI}/audit/records",
            params={
                "since": since,
                "cursor": cursor,
                "limit": limit,
            },
        )

        # Record the cursor, if set.
        cursor = result.body.get("next_cursor", None)

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=result.body.get("records", []))



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Provides helpers for plugin loading."""

from importlib.metadata import EntryPoint, entry_points
from typing import Any

from grove.exceptions import ConfigurationException


def lookup_handler(name: str, group: str) -> EntryPoint:
    """Attempts to locate requested plugin handler.

    This utilises setuptools entrypoints to allow handlers to register themselves for
    use by Grove.

    :param name: The name of the handler to load (e.g. 'aws_ssm').
    :param group: The group the handler belongs to (e.g. 'grove.outputs').

    :raises ConfigurationException: The specified handler could not be located.
    """
    # The 'group' kwarg for entry_points was added in Python 3.10, in order to support
    # older versions of Python 3 we will not be using this feature. Additionally, using
    # get() will raise deprecation warnings to use select() instead. However, select()
    # was also added in Python 3.10, which would break backwards compatibility...
    eps = entry_points()

    for candidate in eps.get(group, []):
        if candidate.name == name:
            return candidate

    raise ConfigurationException(
        f"Requested handler could not be found with name '{name}' (group '{group}')"
    )


def load_handler(name: str, group: str, *args: Any, **kwargs: Any) -> Any:
    """Attempts to locate and load the requested plugin handler.

    This is a convenience method which wrappers the lookup operation, and performs the
    load and instantiation required to hydrate a 'real' handler. Any additional
    arguments passed to load_handler are pass through to the handler during creation.

    :param name: The name of the handler to load (e.g. 'aws_ssm').
    :param group: The group the handler belongs to (e.g. 'grove.outputs').
    """
    cls = lookup_handler(name, group).load()

    return cls(*args, **kwargs)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
from grove.__about__ import __copyright__, __title__, __version__

# -- Project information -----------------------------------------------------

author = "HashiCorp Security (TDR)"
project = __title__.title()
copyright = __copyright__

# The full version, including alpha/beta/rc tags
release = __version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
    "sphinx.ext.viewcode",
]

add_function_parentheses = False
add_module_names = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"
html_title = f"Grove v{__version__}"
# html_logo = "static/grove.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["static"]

html_css_files = [
    "custom.css",
]

html_theme_options = {
    "sidebar_hide_name": False,
    "source_repository": "https://github.com/hashicorp-forge/grove",
    "source_branch": "main",
    "source_directory": "docs/",
    # CSS.
    "light_css_variables": {
        "color-foreground-secondary": "#444",
        "admonition-font-size": "0.9rem",
        "admonition-title-font-size": "0.9rem",
    },
    "dark_css_variables": {
        "color-foreground-secondary": "#9ca0a5",
        "admonition-font-size": "0.9rem",
        "admonition-title-font-size": "0.9rem",
    },
}



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Test connector for Grove."""

import time
from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.constants import CHRONOLOGICAL
from grove.exceptions import NotFoundException

DATESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"


class Connector(BaseConnector):
    NAME = "local_heartbeat"
    POINTER_PATH = "timestamp"
    LOG_ORDER = CHRONOLOGICAL

    @property
    def count(self):
        """Fetches the count of heartbeat messages to emit from the configuration.

        :return: The number of heartbeat messages to emit.
        """
        try:
            return int(self.configuration.count)
        except (AttributeError, ValueError):
            return 5

    @property
    def interval(self):
        """Fetches the interval to emit a heartbeat message from the configuration.

        :return: The heartbeat interval, in seconds.
        """
        try:
            return int(self.configuration.interval)
        except (AttributeError, ValueError):
            return 1

    def collect(self):
        """Generates test log entries at the configured interval."""
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.utcnow() - timedelta(days=7)).isoformat()

        for _ in range(self.count):
            # Writes this batch of entries to the output and updates the pointer.
            self.save(
                [
                    {
                        "type": "heartbeat",
                        "timestamp": datetime.utcnow().strftime(DATESTAMP_FORMAT),
                    }
                ]
            )
            time.sleep(self.interval)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove local file path output handler."""

import datetime
import os
from typing import Optional

from pydantic import Field

from grove.constants import DATESTAMP_FORMAT
from grove.exceptions import AccessException
from grove.outputs import BaseOutput

OBJECT_PATH = (
    "{descriptor}{connector}/{identity}/{year}/{month}/{day}/"
    "{operation}-{datestamp}.{part}{kind}"
)


class Handler(BaseOutput):
    class Configuration(BaseOutput.Configuration):
        """Defines environment variables used to configure the local file handler.

        This should also include any appropriate default values for fields which are not
        required.
        """

        path: str = Field(
            description="The path to the directory to write collected logs to.",
        )

        class Config:
            """Allow environment variable override of configuration fields.

            This also enforce a prefix for all environment variables for this handler.
            As an example the field `path` would be set using the environment variable
            `GROVE_OUTPUT_LOCAL_FILE_PATH`.
            """

            env_prefix = "GROVE_OUTPUT_LOCAL_FILE_"
            case_insensitive = True

    def setup(self):
        """Set up access to local filesystem path.

        This also checks that an output directory is configured, and it is initially
        accessible and writable.

        :raises AccessException: There was an issue accessing to the provided file path.
        """
        # Perform a spot check to see if the directory is writable now. Although this
        # can change, we'd like to bail before we collect any data if it's a simple
        # permissions related misconfiguration.
        if not os.path.isdir(self.config.path):
            raise AccessException(
                f"Configured output path '{self.config.path}' does not exist."
            )

        if not os.access(self.config.path, os.W_OK | os.X_OK):
            raise AccessException(
                f"Configured output path '{self.config.path}' is not writable."
            )

    def submit(
        self,
        data: bytes,
        connector: str,
        identity: str,
        operation: str,
        part: int = 0,
        kind: Optional[str] = ".json.gz",
        descriptor: Optional[str] = "logs/",
    ):
        """Persists captured data to a local file path.

        :param data: Log data to write.
        :param connector: Name of the connector which retrieved the data.
        :param identity: Identity the collected data was collect for.
        :param operation: Operation the collected logs are associated with.
        :param part: Number indicating which part of the same log stream this file
            contains data for. This is used to indicate that the logs are from the same
            collection, but have been broken into smaller files for downstream
            processing.
        :param kind: An optional file suffix to use for files written.
        :param descriptor: An optional path to append to the beginning of the output
            file path.

        :raises AccessException: An issue occurred when writing data.
        """
        # Append a trailing slash to the descriptor if set - to form a path.
        if descriptor and not descriptor.endswith("/"):
            descriptor = f"{descriptor}/"

        # Each log file is output under a particular hierarchy to assist with sharding
        # and ingestion / finding of log data.
        datestamp = datetime.datetime.utcnow()

        filename = OBJECT_PATH.format(
            part=part,
            connector=connector,
            identity=identity,
            operation=operation,
            year=datestamp.strftime("%Y"),
            month=datestamp.strftime("%m"),
            day=datestamp.strftime("%d"),
            datestamp=datestamp.strftime(DATESTAMP_FORMAT),
            kind=kind,
            descriptor=descriptor,
        )

        # Quick and dirty directory traversal check.
        destination = os.path.abspath(
            os.path.join(self.config.path, os.path.dirname(filename))
        )

        if not destination.startswith(self.config.path):
            raise AccessException(
                f"Generated output filepath '{destination}' is outside of the "
                f"configured output directory '{self.config.path}'."
            )

        # Create the directory structure, if needed, and write the data.
        try:
            os.makedirs(destination, exist_ok=True)

            with open(os.path.join(self.config.path, filename), "wb") as fout:
                fout.write(data)
        except OSError as err:
            raise AccessException(f"Unable to write log data to file: {err}")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Oomnitza API client."""

import logging
from typing import Dict, Optional

import requests

from grove.exceptions import RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://{identity}.oomnitza.com"
API_PAGE_SIZE = 200


class Client:
    def __init__(
        self,
        identity: Optional[str] = None,
        token: Optional[str] = None,
    ):
        """Setup a new client.

        :param identity: The name of the Oomnitza organisation.
        :param token: The Oomnitza API token.
        """
        self.logger = logging.getLogger(__name__)
        self.headers = {
            "content-type": "application/json",
        }
        if token:
            self.headers["Authorization2"] = token

        # We need to push the identity into the URI, so we'll keep track of this.
        self._api_base_uri = API_BASE_URI.format(identity=identity)

    def _get(
        self,
        url: str,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP GET against.
        :param params: HTTP parameters to add to the request.

        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        try:
            response = requests.get(url, headers=self.headers, params=params)
            response.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise RequestFailedException(err)

        return HTTPResponse(headers=response.headers, body=response.json())

    def get_activites(
        self,
        cursor: int = 0,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of signing attempt logs.

        :param cursor: Cursor to use when fetching results.
        :param start_date: The earliest date an event represented as an epoch value.
        :param end_date: The latest date an event represented as an epoch value.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        url = f"{self._api_base_uri}/api/v3/activities"

        result = self._get(
            url,
            params={
                "start_date": start_date,
                "end_date": end_date,
                "limit": str(API_PAGE_SIZE),
                "skip": str(cursor),
            },
        )
        # Keep paging until we run out of results.
        data = result.body

        if len(data) == API_PAGE_SIZE:
            cursor += API_PAGE_SIZE
        else:
            cursor = 0

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=data)  # type: ignore



#!/usr/bin/env python3
# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

from cookiecutter.main import cookiecutter

cookiecutter(
    ".",
    no_input=True,
    extra_context={
        "project_type": "outputs",
        "provider_name": "remote",
        "provider_product": "http",
    },
)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Workday Audit connector for Grove."""

from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.connectors.workday.api import Client
from grove.constants import CHRONOLOGICAL
from grove.exceptions import NotFoundException

DATESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


class Connector(BaseConnector):
    NAME = "workday_activity_logging"
    POINTER_PATH = "requestTime"
    LOG_ORDER = CHRONOLOGICAL

    @property
    def base_url(self):
        """Fetches the Workday unique base url from the configuration.

        This is required as this is the url required by WD for authentication and API
        usage.

        :return: The "base_url" portion of the connector's configuration.
        """
        try:
            return self.configuration.base_url
        except AttributeError:
            return None

    @property
    def client_id(self):
        """Fetches the Workday Client ID from the configuration.

        This is required as this is an additional authentication element required by WD.

        :return: The "client_id" portion of the connector's configuration.
        """
        try:
            return self.configuration.client_id
        except AttributeError:
            return None

    @property
    def client_secret(self):
        """Fetches the Workday Client Secret from the configuration.

        This is required as this is an additional authentication element required by WD.

        :return: The "client_secret" portion of the connector's configuration.
        """
        try:
            return self.configuration.client_secret
        except AttributeError:
            return None

    def collect(self):
        """Collects all logs from the Workday Audit API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(
            base_url=self.base_url,
            identity=self.identity,
            client_id=self.client_id,
            client_secret=self.client_secret,
            refresh_token=self.key,
        )

        # Use the refresh token to generate a temporary access token.
        client.get_access_token()

        # Set cursor
        cursor = 0

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago.
        now = datetime.utcnow().strftime(DATESTAMP_FORMAT)
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.utcnow() - timedelta(days=7)).strftime(
                DATESTAMP_FORMAT
            )

        # Get log data from the upstream API. A "from" and "to" datetime query
        # parameters are required.
        while True:
            log = client.get_activity_logging(
                from_date=self.pointer,
                to_date=now,
                cursor=cursor,
            )

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor  # type: ignore
            if not cursor:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the Oomnitza Audit collector."""

import os
import re
import unittest
from unittest.mock import patch

import responses

from grove.connectors.oomnitza.activities import Connector
from grove.models import ConnectorConfig
from tests import mocks


class OomnitzaAuditTestCase(unittest.TestCase):
    """Implements unit tests for the Oomnitza Activities collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                identity="corp",
                key="testkey",
                name="corp",
                connector="test",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @responses.activate
    def test_collect_pagination(self):
        """Ensure pagination is working as expected."""
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/oomnitza/activities/002.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # The last "page" returns an empty cursor.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/oomnitza/activities/001.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # Check the pointer matches the latest execution_time value, and that the
        # expected number of logs were returned.
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 205)
        self.assertEqual(self.connector.pointer, "1682538024")

    @responses.activate
    def test_collect_no_pagination(self):
        """Ensure collection without pagination is working as expected."""
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/oomnitza/activities/001.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # Ensure only a single value is returned, and the pointer is properly set.
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 5)
        self.assertEqual(self.connector.pointer, "1680895957")

    @responses.activate
    def test_collect_no_results(self):
        """Ensure break accurs when there is no results returned."""
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/oomnitza/activities/003.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 0)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the Zoom Activity Log collector."""

import os
import re
import unittest
from unittest.mock import patch

import responses
from responses import matchers

from grove.connectors.zoom.activities import Connector
from grove.models import ConnectorConfig
from tests import mocks


class ZoomActivitiesTestCase(unittest.TestCase):
    """Implements unit tests for the Zoom Activity Logs collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                identity="client_id",
                key="test_key",
                name="test",
                connector="test",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @responses.activate
    def test_collect_pagination(self):
        """Ensure pagination is working as expected."""
        # set bearer token
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/zoom/client/001.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        # first page has a cursor to the next page
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/zoom/activities/001.json"), "r"
                ).read(),
                "utf-8",
            ),
            match=[matchers.header_matcher({"Authorization": "Bearer testtoken"})],
        )

        # The last page returns no cursor.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/zoom/activities/002.json"), "r"
                ).read(),
                "utf-8",
            ),
            match=[matchers.header_matcher({"Authorization": "Bearer testtoken"})],
        )

        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 31)
        self.assertEqual(self.connector.pointer, "2022-08-23T14:45:54Z")

    @responses.activate
    def test_collect_no_pagination(self):
        """Ensure collection without pagination is working as expected."""
        # set bearer token
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/zoom/client/001.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # add one entry result with no cursor
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/zoom/activities/002.json"), "r"
                ).read(),
                "utf-8",
            ),
            match=[matchers.header_matcher({"Authorization": "Bearer testtoken"})],
        )

        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 1)
        self.assertEqual(self.connector.pointer, "2022-08-23T14:45:54Z")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for parsing helpers."""

import unittest

from grove.helpers import parsing


class ParsingHelpersTestCase(unittest.TestCase):
    """Implements tests for parsing helpers."""

    def test_update_by_path(self):
        """Ensures path updating operates as expected."""
        # Multi-dimension.
        expected_multi = {
            "A": {
                "B": {
                    "C": {
                        "D": {
                            "E": "injected",
                            "deepest": True,
                        },
                        "adjacent": True,
                    }
                }
            },
            "top": True,
        }
        self.assertDictEqual(
            expected_multi,
            parsing.update_path(
                {
                    "A": {"B": {"C": {"D": {"deepest": True}, "adjacent": True}}},
                    "top": True,
                },
                "A.B.C.D.E".split("."),
                "injected",
            ),
        )

        # Replacement.
        expected_replace = {
            "A": {"B": {"C": "replaced"}},
        }
        self.assertDictEqual(
            expected_replace,
            parsing.update_path(
                {"A": {"B": {"C": "initial"}}},
                "A.B.C".split("."),
                "replaced",
            ),
        )

        # Single dimension.
        expected_single = {
            "A": "value",
        }
        self.assertDictEqual(
            expected_single,
            parsing.update_path(
                {},
                "A".split("."),
                "value",
            ),
        )

        # Deletion of nested keys.
        self.assertDictEqual(
            {"A": 1},
            parsing.update_path(
                {"A": 1, "B": {"C": [1, 2, 3], "D": {"E": "F"}}},
                "B".split("."),
                None,
            ),
        )

        # Deletion of deeply nested keys.
        self.assertDictEqual(
            {"A": 1, "B": {"D": {"E": "F"}}},
            parsing.update_path(
                {"A": 1, "B": {"C": [1, 2, 3], "D": {"E": "F"}}},
                "B.C".split("."),
                None,
            ),
        )



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove AWS DynamoDB cache handler."""

import logging
import os
from typing import Any, Dict, Optional

from boto3.session import Session
from botocore.exceptions import BotoCoreError, ClientError
from pydantic import BaseSettings, Field, ValidationError

from grove.caches import BaseCache
from grove.exceptions import (
    AccessException,
    ConfigurationException,
    DataFormatException,
    NotFoundException,
)
from grove.helpers import parsing


class Configuration(BaseSettings):
    """Defines environment variables used to configure the AWS DynamoDB handler.

    This should also include any appropriate default values for fields which are not
    required.
    """

    table: str = Field(
        default="grove",
        description="The name of the AWS DynamoDB table to use for the cache.",
    )
    url: Optional[str] = Field(
        description="An optional URL to use when connecting to AWS DynamoDB.",
        default=None,
    )
    assume_role_arn: Optional[str] = Field(
        description="An optional AWS role to assume when authenticating with AWS.",
        default=None,
    )
    table_region: Optional[str] = Field(
        description="The region that the DynamoDB table exists in (default us-east-1)",
        default=os.environ.get("AWS_REGION", "us-east-1"),
    )

    class Config:
        """Allow environment variable override of configuration fields.

        This also enforce a prefix for all environment variables for this handler. As
        an example the field `assume_role_arn` would be set using the environment
        variable `GROVE_CACHE_AWS_DYNAMODB_ASSUME_ROLE_ARN`.
        """

        env_prefix = "GROVE_CACHE_AWS_DYNAMODB_"
        case_insensitive = True


class Handler(BaseCache):
    """This cache handler allows Grove to write objects into an AWS DynamoDB cache."""

    def __init__(self):
        """Sets up access to DynamoDB

        This handler also attempt to assume a configured role in order to allow
        cross-account use - if required.

        :raises ConfigurationException: There was an issue with configuration.
        :raises AccessException: An issue occurred when accessing DynamoDB.
        """
        self.logger = logging.getLogger(__name__)

        # Wrap validation errors to keep them in the Grove exception hierarchy.
        try:
            self.config = Configuration()
        except ValidationError as err:
            raise ConfigurationException(parsing.validation_error(err))

        # Explicit calls to session are mostly used to allow mocks during testing.
        session = Session()
        session_arguments = {}

        if self.config.url:
            self.logger.debug(
                "Using DynamoDB URL from environment variables.",
                extra={"url": self.config.url},
            )
            session_arguments["endpoint_url"] = self.config.url

        try:
            if not self.config.assume_role_arn:
                self._store = session.client(
                    "dynamodb",
                    region_name=self.config.table_region,
                    **session_arguments,
                )
            else:
                self.logger.debug(
                    "Config handler is attempting to assume AWS role for SSM",
                    extra={"role_arn": self.config.assume_role_arn},
                )

                # We don't use session arguments for STS calls. Though perhaps there is
                # a situation where this is needed...?
                sts = session.client("sts")

                role = sts.assume_role(
                    RoleArn=self.config.assume_role_arn,
                    RoleSessionName="GroveDynamoDBHandler",
                )
                self._store = session.client(
                    "dynamodb",
                    region_name=self.config.table_region,
                    aws_access_key_id=role["Credentials"]["AccessKeyId"],
                    aws_secret_access_key=role["Credentials"]["SecretAccessKey"],
                    aws_session_token=role["Credentials"]["SessionToken"],
                )
        except (ClientError, BotoCoreError, KeyError) as err:
            raise AccessException(f"Cache handler failed to access AWS DynamoDB: {err}")

    def get(self, pk: str, sk: str) -> str:
        """Retrieve an value with the given PK / SK.

        :param pk: Partition key of the value to retrieve.
        :param sk: Sort key of the value to retrieve.

        :raises NotFoundException: No value was found.
        :raises AccessException: An issue occurred when getting the value.

        :return: Value from the cache.
        """
        try:
            response = self._store.get_item(
                TableName=self.config.table,
                Key={"pk": {"S": pk}, "sk": {"S": sk}},
            )
            pointer = response["Item"]["data"]["S"]
        except ClientError as err:
            self.logger.error(
                f"Unable to get value from cache. {err}", extra={"pk": pk, "sk": sk}
            )
            raise AccessException(err)
        except KeyError:
            self.logger.info("No value found in cache", extra={"pk": pk, "sk": sk})
            raise NotFoundException()

        return str(pointer)

    def set(
        self,
        pk: str,
        sk: str,
        value: str,
        not_set: bool = False,
        constraint: Optional[str] = None,
    ):
        """Stores the value for the given key in DynamoDB.

        :param pk: Partition key of the value to store.
        :param sk: Sort key of the value to store.
        :param value: Value to store.
        :param not_set: Specifies whether the value must not already be set in the
            cache for the set to be successful.
        :param constraint: An optional condition to use set operation. If provided,
            the currently cached value must match for the delete to be successful.

        :raises ValueError: An incompatible set of parameters were provided.
        :raises AccessException: An issue occurred when storing the value.
        :raises DataFormatException: The provided constraint was not satisfied.
        """
        options: Dict[str, Any] = {}
        options["ExpressionAttributeValues"] = {":data": {"S": str(value)}}

        if not_set and constraint is not None:
            raise ValueError("A value cannot both have a constraint AND not be set.")

        # Construct an appropriate filter based on input parameters.
        if not_set:
            options["ConditionExpression"] = "attribute_not_exists(#data)"

        if constraint is not None:
            options["ConditionExpression"] = "#data = :constraint"
            options["ExpressionAttributeValues"][":constraint"] = {"S": str(constraint)}

        # Attempt to set the item.
        try:
            self._store.update_item(
                TableName=self.config.table,
                Key={"pk": {"S": str(pk)}, "sk": {"S": str(sk)}},
                UpdateExpression="SET #data = :data",
                ExpressionAttributeNames={"#data": "data"},
                **options,
            )
        except ClientError as err:
            # Handle conditional check failures differently as these may indicate
            # concurrent execution.
            error_type = err.response.get("Error", {}).get("Code", "")

            if error_type == "ConditionalCheckFailedException":
                self.logger.error(
                    "Cache set failed as constraint failed.",
                    extra={
                        "pk": pk,
                        "sk": sk,
                        "value": value,
                        "not_set": not_set,
                        "constraint": constraint,
                    },
                )
                raise DataFormatException(err)

            # For everything else, just raise a generic error.
            self.logger.error(
                f"Unable to set value in cache: {err}",
                extra={
                    "pk": pk,
                    "sk": sk,
                    "value": value,
                    "not_set": not_set,
                    "constraint": constraint,
                },
            )
            raise AccessException(err)

    def delete(self, pk: str, sk: str, constraint: Optional[str] = None):
        """Deletes an entry from DynamoDB that has the given PK / SK.

        :param pk: Partition key of the value to delete.
        :param sk: Sort key of the value to delete.
        :param constraint: An optional condition to use during the delete. The value
            provided as the condition must match for the delete to be successful.

        :raises AccessException: An issue occurred when deleting the value.
        :raises DataFormatException: The provided constraint was not satisfied.
        """
        options: Dict[str, Any] = {}

        # Apply a constraint, if set.
        if constraint is not None:
            options["ConditionExpression"] = "#data = :constraint"
            options["ExpressionAttributeNames"] = {"#data": "data"}
            options["ExpressionAttributeValues"] = {
                ":constraint": {"S": str(constraint)}
            }

        try:
            self._store.delete_item(
                TableName=self.config.table,
                Key={"pk": {"S": pk}, "sk": {"S": sk}},
                **options,
            )
        except ClientError as err:
            # Handle conditional check failures differently as these may indicate
            # concurrent execution.
            error_type = err.response.get("Error", {}).get("Code", "")

            if error_type == "ConditionalCheckFailedException":
                self.logger.error(
                    "Cache set failed as constraint failed.",
                    extra={"pk": pk, "sk": sk, "constraint": constraint},
                )
                raise DataFormatException(err)

            # For everything else, just raise a generic error.
            self.logger.error(
                f"Unable to delete value from cache: {err}",
                extra={"pk": pk, "sk": sk, "constraint": constraint},
            )
            raise AccessException(err)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for connector locking."""

import os
import time
import unittest
from unittest.mock import patch

from grove.connectors import BaseConnector
from grove.constants import ENV_GROVE_LOCK_DURATION
from grove.exceptions import ConcurrencyException
from grove.models import ConnectorConfig
from tests import mocks


class ConnectorLockingTestCase(unittest.TestCase):
    """Implements tests for connector locking."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def test_lock_acquire(self):
        """Ensures locks can be acquired and concurrent executions are prevented."""
        config = ConnectorConfig(
            key="token",
            name="test",
            identity="1FEEDFEED1",
            connector="example_one",
            operation="first",
        )
        context = {
            "runtime": "test_harness",
            "runtime_id": "NA",
        }

        # Acquire a lock.
        first_execution = BaseConnector(config=config, context=context)
        first_execution.lock()

        # Simulate a parallel execution of the same connector instance.
        second_execution = BaseConnector(config=config, context=context)

        # In-memory cache does not support locking as it's only intended for local
        # "one-shot" execution, and development use. As a result, we have to currently
        # clone the state of one cache to the other to simulate this.
        #
        # TODO: Remove the need for this, as it's going to cause confusion in future.
        second_execution._cache._data = first_execution._cache._data

        # Ensure acquiring the lock fails.
        with self.assertRaises(ConcurrencyException):
            second_execution.lock()

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def test_lock_takeover(self):
        """Ensures locks can be taken over if too old."""
        # Override lock duration to something trivial.
        os.environ[ENV_GROVE_LOCK_DURATION] = str("5")

        config = ConnectorConfig(
            key="token",
            name="test",
            identity="1FEEDFEED1",
            connector="example_one",
            operation="first",
        )
        context = {
            "runtime": "test_harness",
            "runtime_id": "NA",
        }

        # Acquire a lock.
        first_execution = BaseConnector(config=config, context=context)
        first_execution.lock()

        # Attempt to acquire a lock for the same connector instance - after the lock
        # has expired. This should succeed due to expiry.
        second_execution = BaseConnector(config=config, context=context)

        # In-memory cache does not support locking as it's only intended for local
        # "one-shot" execution, and development use. As a result, we have to currently
        # clone the state of one cache to the other to simulate this.
        #
        # TODO: Remove the need for this, as it's going to cause confusion in future.
        second_execution._cache._data = first_execution._cache._data

        # Ensure acquiring the lock fails before expiration.
        with self.assertRaises(ConcurrencyException):
            second_execution.lock()

        # Wait to ensure the lock has expired.
        time.sleep(5)

        # Ensure the lock is taken-over successfully after expiration.
        second_execution.lock()

        # Ensure subsequent attempts of the first fail due to this lock takeover.
        with self.assertRaises(ConcurrencyException):
            first_execution.lock()

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def test_lock_unlock(self):
        """Ensures unlock functions as expected."""
        # Override lock duration to something trivial.
        os.environ[ENV_GROVE_LOCK_DURATION] = str("10")

        config = ConnectorConfig(
            key="token",
            name="test",
            identity="1FEEDFEED1",
            connector="example_one",
            operation="first",
        )
        context = {
            "runtime": "test_harness",
            "runtime_id": "NA",
        }

        # Acquire a lock.
        first_execution = BaseConnector(config=config, context=context)
        first_execution.lock()

        # Attempt to acquire a lock for the same connector instance - after the lock
        # has expired. This should succeed due to expiry.
        second_execution = BaseConnector(config=config, context=context)

        # In-memory cache does not support locking as it's only intended for local
        # "one-shot" execution, and development use. As a result, we have to currently
        # clone the state of one cache to the other to simulate this.
        #
        # TODO: Remove the need for this, as it's going to cause confusion in future.
        second_execution._cache._data = first_execution._cache._data

        # Ensure acquiring the lock fails before expiration.
        with self.assertRaises(ConcurrencyException):
            second_execution.lock()

        # Unlock the first connector, and then ensure the second calling lock does not
        # error. First though, we need to clone the newly updated cache again.
        #
        # TODO: Remove the need for this, as it's going to cause confusion in future.
        first_execution.unlock()
        second_execution._cache._data = first_execution._cache._data
        second_execution.lock()



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Provides helpers for parsing."""

import json
import re
from typing import Any, Dict, List

from pydantic import ValidationError


def validation_error(exc: ValidationError):
    """Parse Pydantic validation exceptions into a user readable string.

    :param exc: The Pydantic ValidationError to parse.

    :return: The exception as a string, including fields with validation errors.
    """
    message = "Handler configuration is not valid"
    try:
        prefix = exc.model.Config.env_prefix  # type: ignore
    except AttributeError:
        prefix = ""

    # Ensure the validation errors are included in the logged error message.
    for error in exc.errors():
        field = str(error["loc"][0]).upper()
        problem = error["msg"]

        # Add the environment variable prefix onto the field name.
        message = f"{message}, {prefix}{field} {problem}"

    return message


def quick_copy(value: Any):
    """Performs a quick deep copy by marshalling and unmarshalling to JSON.

    This operation, although strange, is significantly quicker than copy.deepcopy().
    This has been moved into a helper to enable potential performance improvements in
    future without code changes in processors being required.

    :param value: The value to perform a deep copy of.

    :return: The deep copy of the input value.
    """
    return json.loads(json.dumps(value))


def quote_aware_split(value: str, delimiter=".") -> List[str]:
    """Splits a value by delimiter, returning a list.

    This function is quote aware, ensuring that splitting will not occur inside of a
    value quoted with single-quotes.

    :param value: The value to split.
    :param delimiter: The delimiter to split using.

    :return: A list of elements split from the input value.
    """
    fields = []

    for field in re.split(rf"({re.escape(delimiter)}|'.*?')", value):
        # Drop empty and delimiter only fields.
        field = field.strip(delimiter)
        field = field.strip()
        field = re.sub(r"^'(.*)'$", r"\1", field)

        if field:
            fields.append(field)

    return fields


def update_path(
    candidate: Dict[str, Any],
    path: List[str],
    value: Any,
    replace: bool = False,
) -> Dict[str, Any]:
    """Updates or deletes values under the specified path for the provided candidate.

    A path is a list of strings delimited string which express a location within the
    candidate data. If the location is not nested, a single element list should be
    provided.

    As an example of this, a path of `["A", "B", "C"]` expresses that the specified
    value should be set under `{"A": {"B": {"C": value } } }` within the candidate
    dictionary.

    This function recursively walks the provided candidate dictionary until the location
    specified by the path has been located. Once found, the provided value will perform
    on of the following operations:

        1. By default, the provided value will be combined with the existing value.
        2. If `replace` is `True`, the existing value will be replaced with the new.
        3. If `None` is provided as the new value, the specified path will be deleted.

    :param candidate: The dictionary to update.
    :param path: The path to the key to update, as a list of strings.
    :param value: The value to assign to the destination path, or None to delete.
    :param replace: Whether to replace the current value with the new value, or combine.

    :return: The updated dictionary.
    """
    key = path.pop(0)

    # Set the value on the last recursion.
    if len(path) < 1:
        if value is None:
            del candidate[key]
            return candidate

        # If replace is set, don't combine the new value with the existing.
        if replace:
            candidate[key] = value
            return candidate

        # By default, combine the new value with the existing value(s) - making sure to
        # handle dictionaries as well as lists.
        if key in candidate and isinstance(candidate[key], list):
            candidate[key].append(value)
        else:
            candidate = {**candidate, key: value}

        return candidate

    # If recursing, ensure the child we're trying to recurse into exists.
    if key not in candidate:
        candidate[key] = {}

    candidate[key] = update_path(candidate[key], path, value, replace=replace)

    return candidate



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the Okta API Client."""

import unittest

from grove.connectors.okta.api import Client


class OktaAuditTestCase(unittest.TestCase):
    """Implements unit tests for the Okta API Client."""

    def test_client_parse_link_header(self):
        """Ensure Link header parsing works as expected."""
        client = Client(identity="example", token="Key")

        # Ensure the next page URL is extracted correctly.
        candidate = (
            '<https://example.okta.com/api/v1/logs?since=2021&limit=1000&sortOrder=ASCENDING&after=1625253400145_1>; rel="next", '
            '<https://example.okta.com/api/v1/logs?since=2021&limit=1000&sortOrder=ASCENDING&after=1625251850961_1>; rel="self"'
        )
        expected = "https://example.okta.com/api/v1/logs?since=2021&limit=1000&sortOrder=ASCENDING&after=1625253400145_1"
        self.assertEqual(client._parse_link_header(candidate), expected)

        # Ensure its the last page
        candidate = (
            '<https://example.okta.com/api/v1/logs?since=2021&limit=1000&sortOrder=ASCENDING&after=1625253400145_1>; rel="next", '
            '<https://example.okta.com/api/v1/logs?since=2021&limit=1000&sortOrder=ASCENDING&after=1625253400145_1>; rel="self"'
        )
        with self.assertRaises(ValueError):
            client._parse_link_header(candidate)

        # Ensure SSRF mitigation is working correctly.
        candidate = (
            '<https://192.0.2.1:8080/api/v1/logs?since=2021&limit=1000&sortOrder=ASCENDING&after=1625253400145_1>; rel="next", '
            '<https://example.okta.com/api/v1/logs?since=2021&limit=1000&sortOrder=ASCENDING&after=1625251850961_1>; rel="self"'
        )
        with self.assertRaises(ValueError):
            client._parse_link_header(candidate)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for the local memory cache handler."""

import unittest

from grove.caches.local_memory import Handler
from grove.exceptions import DataFormatException


class LocalMemoryCacheTestCase(unittest.TestCase):
    """Implements tests for the local memory cache handler."""

    def test_set(self):
        """Ensures our set operations match expectations."""
        handler = Handler()

        # Set a value with no constraints.
        handler.set("test", "fixture", "0000")

        # Ensure not_set is working as expected.
        handler.set("test", "fixture", "0001", not_set=False)
        with self.assertRaises(DataFormatException):
            handler.set("test", "fixture", "0002", not_set=True)

        # Ensure constraints are working as expected.
        handler.set("test", "fixture", "0003", constraint="0001")
        with self.assertRaises(DataFormatException):
            handler.set("test", "fixture", "0004", constraint="____")

        # Ensure constraints and not_set are working together.
        handler.set("test", "fixture", "0005", constraint="0003", not_set=False)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for the local file secrets backend."""

import os
import tempfile
import unittest

from grove.secrets.local_file import Handler


class SecretsLocalFileTestCase(unittest.TestCase):
    """Implements tests for the local file secrets backend."""

    def setUp(self):
        self.fixtures = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "fixtures/")
        )

    def test_relative_path(self):
        """Ensures a secret can be read from a relative file path."""
        expected = "_Super_S3cret_Stuff."

        with tempfile.NamedTemporaryFile("w") as fout:
            fout.write(expected)
            fout.write("\n")
            fout.flush()

            # Validate loading with a path prefix.
            os.environ["GROVE_SECRET_LOCAL_FILE_PATH_PREFIX"] = os.path.dirname(
                fout.name
            )

            self.secrets = Handler()
            self.assertEqual(self.secrets.get(os.path.basename(fout.name)), expected)

    def test_absolute_path(self):
        """Ensures a secret can be read from an absolute file path."""
        expected = "_Super_S3cret_Stuff."

        with tempfile.NamedTemporaryFile("w") as fout:
            fout.write(expected)
            fout.write("\n")
            fout.flush()

            # Validate loading with a path prefix.
            self.secrets = Handler()
            self.assertEqual(self.secrets.get(fout.name), expected)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove AWS SSM parameter store secret handler."""

import logging
import os
from typing import Optional

import jmespath
from boto3.session import Session
from botocore.exceptions import BotoCoreError, ClientError
from pydantic import BaseSettings, Field, ValidationError

from grove.exceptions import AccessException, ConfigurationException
from grove.helpers import parsing
from grove.secrets import BaseSecret


class Configuration(BaseSettings):
    """Defines environment variables used to configure the AWS SSM handler.

    This should also include any appropriate default values for fields which are not
    required.
    """

    assume_role_arn: Optional[str] = Field(
        description="An optional AWS role to assume when authenticating with AWS.",
        default=None,
    )
    ssm_region: Optional[str] = Field(
        description="The region that the parameter store exists in (default us-east-1)",
        default=os.environ.get("AWS_REGION", "us-east-1"),
    )

    class Config:
        """Allow environment variable override of configuration fields.

        This also enforce a prefix for all environment variables for this handler. As
        an example the field `assume_role_arn` would be set using the environment
        variable `GROVE_SECRET_AWS_SSM_ASSUME_ROLE_ARN`.
        """

        env_prefix = "GROVE_SECRET_AWS_SSM_"
        case_insensitive = True


class Handler(BaseSecret):
    """A configuration handler to read secrets from AWS SSM."""

    def __init__(self):
        """Sets up access to AWS SSM.

        This handler also attempt to assume a configured role in order to allow
        cross-account use - if required.

        :raises ConfigurationException: There was an issue with configuration.
        :raises AccessException: An issue occurred when accessing SSM.
        """
        self.logger = logging.getLogger(__name__)

        # Wrap validation errors to keep them in the Grove exception hierarchy.
        try:
            self.config = Configuration()
        except ValidationError as err:
            raise ConfigurationException(parsing.validation_error(err))

        # Explicit calls to session are mostly used to allow mocks during testing.
        session = Session()

        try:
            if not self.config.assume_role_arn:
                self._ssm = session.client("ssm", region_name=self.config.ssm_region)
            else:
                sts = session.client("sts")
                self.logger.debug(
                    "Secrets handler is attempting to assume AWS role for SSM",
                    extra={"role_arn": self.config.assume_role_arn},
                )

                role = sts.assume_role(
                    RoleArn=self.config.assume_role_arn,
                    RoleSessionName="GroveSecretHandler",
                )
                self._ssm = session.client(
                    "ssm",
                    region_name=self.config.ssm_region,
                    aws_access_key_id=role["Credentials"]["AccessKeyId"],
                    aws_secret_access_key=role["Credentials"]["SecretAccessKey"],
                    aws_session_token=role["Credentials"]["SessionToken"],
                )
        except (ClientError, BotoCoreError, KeyError) as err:
            raise AccessException(
                f"Secrets handler was unable to access AWS SSM: {err}"
            )

    def get(self, id: str) -> str:
        """Gets and returns an encrypted parameter from AWS SSM.

        :param id: The path of the secret to retrieve.

        :return: The decrypted and plain-text secret.
        """
        try:
            parameter = self._ssm.get_parameter(Name=id, WithDecryption=True)
        except (ClientError, BotoCoreError) as err:
            raise AccessException(
                f"Secrets handler failed to read secret from AWS SSM path {id}: {err}"
            )

        return jmespath.search("Parameter.Value", parameter)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Oomnitza Activities connector for Grove."""
import time
from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.connectors.oomnitza.api import Client
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "oomnitza_activities"
    POINTER_PATH = "timestamp"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    def collect(self):
        """Collects Oomnitza activities from the Oomnitza API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(token=self.key, identity=self.identity)

        # Set cursor
        cursor = 0

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to 2 days ago. In the case of the Oomnitza activities API the pointer
        # is the value of the "timestamp" field from the latest record retrieved from
        # the API - which is in epoch. The Oomnitza API doesnt account for milliseconds.
        now = datetime.fromtimestamp(time.time()).strftime("%s")

        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (
                datetime.fromtimestamp(time.time()) - timedelta(days=2)
            ).strftime("%s")

        # Get log data from the upstream API. A "start_date" and "end_date" datetime
        # query parameters are required.
        while True:
            log = client.get_activites(
                start_date=self.pointer, end_date=now, cursor=cursor
            )

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor  # type: ignore
            if not cursor:
                break



"""{{ cookiecutter.project_description }}."""

from typing import Any, Dict, List

from pydantic import Extra

from grove.models import ProcessorConfig
from grove.outputs import BaseOutput


class Handler(BaseOutput):
    """A Grove handler for {{ cookiecutter.provider_name }} {{ cookiecutter.provider_product }}."""

    class Configuration(ProcessorConfig, extra=Extra.forbid):
        """Expresses the configuration and associated validators for the processor."""

        pass

    def process(self, entry: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Performs a set of processing operations.

        :param entry: A collected log entry.

        :return: The processed log entry with fields mapped, as a list.
        """
        return [entry]



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the path zipping processor."""

import json
import os
import unittest
from unittest.mock import patch

from grove.models import ProcessorConfig
from grove.processors import zip_paths
from tests import mocks


class ProcessorZipPathsTestCase(unittest.TestCase):
    """Implements unit tests for the path zipping processor."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Setup the processor and associated configuration for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))

        # Create a mapping compatible with the Okta system log fixture.
        self.processor = zip_paths.Handler(
            ProcessorConfig(
                name="Zip parameters",
                processor="zip_paths",
                source="events.parameters",
                key="name",
                values=[
                    "value",
                    "boolValue",
                    "multiValue",
                ],
            )
        )

    def test_zip_paths(self):
        """Ensure path zipping operates as expected."""
        # Load and process the fixture.
        entries = json.load(
            open(
                os.path.join(self.dir, "fixtures/gsuite/activities/003.json"),
                "r",
            )
        )

        # Strip the outer items list from the input logs, as Grove will have stripped
        # this already.
        entries = entries.get("items", [])

        # Modify the entry to drop the list of entries, in favour of a single element.
        # This simulates Google Workspace / GSuite activity logs being run through the
        # path splitter first.
        entry = entries[0]
        entry["events"] = entry["events"][2]

        # Process the entries.
        records = []
        records.extend(self.processor.process(entry))

        # Confirm fields are remapped into key / values after processing.
        record = records[0]
        self.assertEqual(record["events"]["parameters"]["owner_is_shared_drive"], False)
        self.assertEqual(record["events"]["parameters"]["visibility"], "private")
        self.assertEqual(record["events"]["parameters"]["new_value"], ["owner"])



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""GitHub API client.

As PyGitHub does not currently support the Audit API, this client has been created in
the interim.
"""

import datetime
import logging
import time
from typing import Dict, Optional
from urllib.parse import unquote

import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse


class Client:
    def __init__(
        self,
        hostname: str = "api.github.com",
        scope: str = "orgs",
        identity: Optional[str] = None,
        token: Optional[str] = None,
        retry: Optional[bool] = True,
    ):
        """Setup a new GitHub audit log client.

        :param hostname: Hostname of the GitHub API to interact with.
        :param identity: Name of the GitHub identity to collect audit logs for.
        :param token: Personal Access Token (PAT) to authenticate with.
        :param scope: The scope of the user - such as "orgs" or "enterprises".
        :param retry: Automatically retry if recoverable errors are encountered, such as
            rate-limiting.
        """
        self.retry = retry
        self.scope = scope
        self.hostname = hostname
        self.identity = identity
        self.logger = logging.getLogger(__name__)
        self.headers = {
            "Accept": "application/vnd.github.v3+json",
            "Authorization": f"Bearer {token}",
        }

    def _parse_link_header(self, link: str) -> str:
        """Attempt to parse the "next" URL from a provided Link header.

        :param link: Value of a Link header returned from a previous request.

        :raises ValueError: No, or an invalid, Link header was encountered.

        :return: Extracted "Next" URL from the provided Link header.
        """
        # A link header may contain N entries ("first", "next", and "last").
        url = None
        links = link.split(",")

        for entry in links:
            link_parts = entry.split(";")
            link_rel = link_parts[-1].strip()

            # There may not always be 'next' - such as in the case of the LAST page of
            # results.
            if "next" in link_rel:
                url = unquote(link_parts[0].strip().lstrip("<").rstrip(">"))

        # Very likely the last page.
        if not url:
            raise ValueError()

        # Try to mitigate SSRFs where a baked Link header is returned.
        if self.hostname.lower() not in url.lower():
            raise ValueError(
                f"{self.hostname} not found in Link header ({url}). Ignoring."
            )

        return url

    def _get(
        self,
        url: str,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP GET against.
        :param params: HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        while True:
            try:
                response = requests.get(url, headers=self.headers, params=params)
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as err:
                # 403s OR 429s are used with a header to indicate rate-limit exceeded.
                if int(getattr(err.response, "status_code", 0)) not in [403, 429]:
                    raise RequestFailedException(err)

                if err.response.headers.get("X-RateLimit-Remaining") != "0":
                    raise RequestFailedException(err)

                # Retry on rate-limit, but only if requested.
                self.logger.warning("Rate-limit was exceeded during request")
                if not self.retry:
                    raise RateLimitException(err)

                time_current = int(
                    datetime.datetime.now(datetime.timezone.utc).strftime("%s")
                )
                time_ratelimit_reset = int(
                    err.response.headers.get("X-RateLimit-Reset", time_current)
                )
                time_wait = time_ratelimit_reset - time_current

                # If the rate-limit retry is greater than a few of minutes, just bail as
                # we'll pick back up at the next execution.
                if time_wait >= 180:
                    raise RateLimitException(err)

                # Only wait for a second if the retry time was unset or in the past.
                if time_wait > 0:
                    time.sleep(time_wait)
                else:
                    time.sleep(1)

        return HTTPResponse(headers=response.headers, body=response.json())

    def get_audit_log(
        self,
        phrase: Optional[str] = None,
        include: Optional[str] = "all",
        order: Optional[str] = "asc",
        cursor: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of audit logs which match the requested filter and event type.

        :param phrase: Search phrase to use when filtering logs.
        :param include: Event type to return (web, git, or all).
        :param order: Order to return results (asc, desc).
        :param cursor: Cursor to use when fetching results. Supersedes other parameters.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        # Use the cursor URL if set, otherwise construct the initial query.
        if cursor is not None:
            self.logger.debug(
                "Collecting next page with provided cursor", extra={"cursor": cursor}
            )
            result = self._get(cursor)
        else:
            # See psf/requests issue #2651 for why we can happily pass in None values
            # and not have the request key added to the URI.
            result = self._get(
                f"https://{self.hostname}/{self.scope}/{self.identity}/audit-log",
                params={
                    "phrase": phrase,
                    "include": include,
                    "order": order,
                    "per_page": "100",
                },
            )

        # Track the results.
        try:
            cursor = self._parse_link_header(result.headers.get("Link", ""))
        except ValueError:
            cursor = None

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=result.body)  # type: ignore



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Data models used throughout Grove."""

import base64
import binascii
from enum import Enum
from typing import Dict, List

from pydantic import BaseModel, Extra, Field, root_validator, validator

from grove.constants import OPERATION_DEFAULT
from grove.exceptions import DataFormatException


def decode(value: str, encoding: str) -> str:
    """Decode a value using the specified encoding.

    :param value: The encoded value to decode.
    :param encoding: The encoding the value is encoded with.

    :raises DataFormatException: Decoding failed due to an issue with the data, or an
        due to an unsupported encoding method.

    :return: The decoded value.
    """
    # TODO: Decoded binary data would be cast to string here, this could present a
    # for a future unknown plugin or connector?
    try:
        if encoding == "base64" and value:
            return str(base64.b64decode(value), "utf-8")
    except binascii.Error as err:
        raise DataFormatException(f"Unable to base64 decode data, {err}")

    # More formats may be supported later.
    raise DataFormatException(f"Unknown encoding method '{encoding}'")


class ProcessorConfig(BaseModel, extra=Extra.allow):
    """A processor configuration object.

    A processor configuration object represents information used by processors to
    perform some set of operations on log entries. This base configuration object
    is bare-bones as processors may define their own required configuration fields.
    """

    # Name is an arbitrary name administrators can provide to processors to enable
    # better tracking and identification of processors.
    name: str

    # Processor defines the processor which should be run. This must match the plugin
    # entrypoint name.
    processor: str


class OutputStream(str, Enum):
    """Defines supported output 'streams'.

    This is used to allow routing of original / raw collected data differently to
    post processed data.
    """

    raw = "raw"
    processed = "processed"


class ConnectorConfig(BaseModel, extra=Extra.allow):
    """Defines the connector configuration structure.

    A configuration object represents information which Grove uses to call a given
    connector. All connectors must have at least a name, key, identity, and connector
    defined.

    Connector configuration objects support the addition of arbitrary fields to enable
    service specific authentication and configuration concerns to be defined. This is
    useful for services which require more than one factor for authentication, or other
    tuneables such as self-hosted API instance FQDN.
    """

    name: str
    identity: str
    connector: str

    # Although key is required, it may be set via secret reference. This is checked
    # during validation.
    key: str = Field("")

    # Allow the connector to be disabled via configuration flag.
    disabled: bool = Field(False)

    # Secrets is used to mark which fields are considered to be secrets, and their
    # associated location in the configured secrets backend.
    secrets: Dict[str, str] = Field({})

    # Similar to secrets, Encoding is used to mark fields which are encoded due in some
    # form which must be decoded before use. This is often used for base64 encoding
    # binary data or nested JSON.
    encoding: Dict[str, str] = Field({})

    # Operations allow connectors and users to filter which 'type' of events to collect
    # from API endpoints which allow filtering records to return.
    operation: str = Field(OPERATION_DEFAULT)

    # Processors allow processing of data during collection.
    processors: List[ProcessorConfig] = Field([])

    # Outputs allows specification of what type of data to output, and with what
    # descriptor. By default, any processed logs will be output with a descriptor of
    # 'processed', and raw logs with a descriptor of 'logs'.
    outputs: Dict[str, OutputStream] = Field(
        {
            "logs": OutputStream.raw,
            "processed": OutputStream.processed,
        }
    )

    @validator("key")
    def _validate_key_or_secret(cls, value, values, field):  # noqa: B902
        """Ensures that 'key' is set directly or a reference is present in 'secrets'.

        This is used to ensure that a key is always set, whether directly, or will be
        fetched from the configured secrets backend at runtime (as indicated by an entry
        in the secrets field).
        """
        if value is None and field not in values["secrets"]:
            raise ValueError(f"Required field '{field}' is missing")

        return value

    @root_validator(pre=True)
    def _decode_fields(cls, values):  # noqa: B902
        """Automatically decode fields using the specified encoding during data loading.

        If a field is listed in both the 'secrets' field and this 'encoding' field,
        decoding will be deferred to after secrets have been retrieved. This allows
        encoding of externally stored secrets to be specified using this field.

        This is intended to allow fields which contain data that is not easily expressed
        in JSON to be encoded with an appropriate scheme. As an example, this is useful
        to allow binary or multi-line data by first encoding the value with base64.

        Other encoding schemes may be supported in future, but for now only base64 is
        supported.
        """
        for field, encoding in values.get("encoding", {}).items():
            # If the secret is externally stored decoding will be performed after the
            # secret has been retrieved. Right now, this field should not exist as it
            # won't have been fetched yet.
            if field in values.get("secrets", {}):
                continue

            values[field] = decode(values.get(field), encoding)

        return values



"""{{ cookiecutter.provider_name }} {{ cookiecutter.provider_product }} example logs connector for Grove."""

from datetime import datetime, timedelta, timezone

from grove.connectors import BaseConnector
from grove.constants import CHRONOLOGICAL
from grove.exceptions import NotFoundException
from grove.types import AuditLogEntries


class Connector(BaseConnector):
    NAME = "{{ cookiecutter.provider_name }}_{{ cookiecutter.provider_product }}_example_logs"
    LOG_ORDER = CHRONOLOGICAL
    POINTER_PATH = "timestamp"

    @property
    def optional_setting(self):
        """Fetches an optional setting from the connector configuration document.

        :return: The "optional_setting" component of the connector configuration.
        """
        try:
            return self.configuration.optional_setting
        except AttributeError:
            return "Some Default value"

    def collect(self):
        """Collects example logs from the {{ cookiecutter.provider_name }} API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        # TODO: Setup API client for product.

        try:
            _ = self.pointer
        except NotFoundException:
            # TODO: Change the default pointer value to the correct format.
            self.pointer = str(
                int((datetime.now(timezone.utc) - timedelta(days=7)).timestamp())
            )

        # TODO: Retrieve logs from API in pages in a loop, calling self.save() on each
        # page to reduce memory use, and allow partial collection.
        while True:
            # TODO: Replace with call to API.
            log = AuditLogEntries(None, [])

            # Save this batch of log entries.
            self.save(log.entries)

            # Break out of loop when there are no more pages.
            cursor = log.cursor  # type: ignore
            if cursor is None:
                break



"""{{ cookiecutter.project_description }}."""

import logging
from typing import Optional

from pydantic import BaseSettings, Field, ValidationError

from grove.configs import BaseConfig
from grove.exceptions import ConfigurationException
from grove.helpers import parsing


class Configuration(BaseSettings):
    """Defines environment variables used to configure the {{ cookiecutter.provider_name }} {{ cookiecutter.provider_product }} handler.

    This should also include any appropriate default values for fields which are not
    required.
    """

    required_setting: str = Field(
        default="thing",
        description="A required configuration value (if unset a default will be used)",
    )
    optional_setting: Optional[str] = Field(
        description="An optional configuration value."
    )

    class Config:
        """Allow environment variable override of configuration fields.

        This also enforce a prefix for all environment variables for this handler. As
        an example the field `required_setting` would be set using the environment
        variable `{{ cookiecutter.project_slug.upper() }}_REQUIRED_SETTING`.
        """

        env_prefix = "{{ cookiecutter.project_slug.upper() }}_"
        case_insensitive = True


class Handler(BaseConfig):
    """A Grove handler for {{ cookiecutter.provider_name }} {{ cookiecutter.provider_product }}."""

    def __init__(self):
        """Sets up access to {{ cookiecutter.provider_product }}

        :raises ConfigurationException: There was an issue with configuration.
        """
        self.logger = logging.getLogger(__name__)

        # Wrap validation errors to keep them in the Grove exception hierarchy.
        try:
            self.config = Configuration()
        except ValidationError as err:
            raise ConfigurationException(parsing.validation_error(err))



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""SalesForce Marketing Cloud Security security event connector for Grove."""

from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.connectors.sfmc.api import Client
from grove.constants import CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "sfmc_security_events"
    POINTER_PATH = "createdDate"
    LOG_ORDER = CHRONOLOGICAL

    def collect(self):
        """Collects all logs from the SFMC API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(identity=self.identity, token=self.key)
        cursor = 1

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago.
        now = datetime.utcnow()
        isoformat = {"sep": "T", "timespec": "milliseconds"}
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (now - timedelta(days=7)).isoformat(**isoformat)

        # Get log data from the upstream API. A "enddate" argument is used here as the
        # API specification simply says that "the default is today". It is unclear
        # whether this is midnight today, or "now".
        while True:
            log = client.get_audit_events(
                cursor=cursor,
                startdate=self.pointer,
                enddate=now.isoformat(**isoformat),
            )

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor  # type: ignore
            if not cursor:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Provides Grove configuration storage using supported backends."""

import abc
from typing import List

from grove.models import ConnectorConfig


class BaseConfig(abc.ABC):
    @abc.abstractmethod
    def get(self, id: str) -> List[ConnectorConfig]:
        """Gets and returns one or more connector configuration objects.

        Multiple connector configurations may be returned by backends which provide a
        recursive option where a less specific identifier is provided. This is used to
        enable lookup of all configuration documents under a "path" inside of the given
        backend - such as Consul, or AWS SSM.

        :param id: The identifier to use when querying for connector configuration.

        :return: A list of ConnectorConfig objects.
        """
        pass



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Terraform Cloud audit trails connector for Grove."""

from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.connectors.tfc.api import Client
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "tfc_audit_trails"
    POINTER_PATH = "timestamp"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    def collect(self):
        """Collects TFC audit trail from the TFC audit-trails API.

        https://www.terraform.io/docs/cloud/api/audit-trails.html

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(token=self.key)
        cursor = 1

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago. In the case of the TFC audit API the pointer is the
        # value of the "timestamp" field from the latest record retrieved from the
        # API - which is in ISO8601 Format.
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.utcnow() - timedelta(days=7)).isoformat()

        # Get log data from the upstream API.
        while True:
            log = client.get_trails(since=self.pointer, cursor=cursor)

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor  # type: ignore
            if not cursor:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for the Base connector."""

import unittest
from unittest.mock import patch

from grove.connectors import BaseConnector
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException
from grove.models import ConnectorConfig
from tests import mocks


class BaseConnectorTestCase(unittest.TestCase):
    """Implements tests for the Base connector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def test_pointers(self):
        """Ensures pointers can be retrieved without issue."""
        first = BaseConnector(
            config=ConnectorConfig(
                key="token",
                name="test",
                identity="1FEEDFEED1",
                connector="example_one",
                operation="first",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )
        second = BaseConnector(
            config=ConnectorConfig(
                key="token",
                name="test",
                identity="1FEEDFEED1",
                connector="example_one",
                operation="second",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

        # Ensure the first pointer being set and saved doesn't interfere with the
        # second - as the operations differ.
        first.pointer = "AAAAAAAA"
        self.assertEqual(first.pointer, "AAAAAAAA")

        # Ensure no pointer is found for the second operation.
        with self.assertRaises(NotFoundException):
            _ = second.pointer

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def test_plain_key(self):
        """Ensures keys are used directly if not set as a secret."""
        first = BaseConnector(
            config=ConnectorConfig(
                key="token",
                name="test",
                identity="1FEEDFEED1",
                connector="example_one",
                operation="second",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )
        self.assertEqual(first.key, "token")

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def test_encoded_key(self):
        """Ensures key encoding is handled."""
        # Ensure keys are base64 decoded properly.
        connector = BaseConnector(
            config=ConnectorConfig(
                key="QUJDREVG",
                name="test",
                identity="1FEEDFEED1",
                connector="example_one",
                operation="second",
                encoding={
                    "key": "base64",
                },
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )
        self.assertEqual(connector.key, "ABCDEF")

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def test_recover_from_incomplete(self):
        # Setup a connector which operates in reverse chronological mode.
        connector = BaseConnector(
            config=ConnectorConfig(
                key="token",
                name="test",
                identity="1FEEDFEED1",
                connector="example_one",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )
        connector.NAME = "test"
        connector.POINTER_PATH = "time"
        connector.LOG_ORDER = REVERSE_CHRONOLOGICAL

        # Simulates logs returned from an upstream / external service.
        logs = [
            {"time": "2022-11-20T20:38:18.382Z", "name": "AAH"},  #   ^
            {"time": "2022-11-20T20:37:18.382Z", "name": "AAG"},  #   | New logs since
            {"time": "2022-11-20T20:36:18.382Z", "name": "AAF"},  #   | the failed
            {"time": "2022-11-20T20:35:18.382Z", "name": "AAE"},  #   | collection.
            {"time": "2022-11-20T20:34:18.382Z", "name": "AAD"},  #   |
            {"time": "2022-11-20T20:33:18.382Z", "name": "AAC"},  # - /
            {"time": "2022-11-20T20:31:18.382Z", "name": "AAB"},  # <-. Most recent log
            {"time": "2022-11-20T20:30:10.772Z", "name": "AAA"},  #   | from failed run.
        ]

        # Setup the cache to match the described failure case.
        connector._cache.set(
            pk="pointer_next.test.06dc0fd3c08a2bc6a33f5460da9fea10",
            sk="all",
            value="2022-11-20T20:31:18.382Z",
        )
        connector._cache.set(
            pk="window_start.test.06dc0fd3c08a2bc6a33f5460da9fea10",
            sk="all",
            value="2022-11-20T20:31:18.382Z",
        )
        connector._cache.set(
            pk="pointer_previous.test.06dc0fd3c08a2bc6a33f5460da9fea10",
            sk="all",
            value="2022-11-20T20:30:10.772Z",
        )
        connector._cache.set(
            pk="window_end.test.06dc0fd3c08a2bc6a33f5460da9fea10",
            sk="all",
            value="2022-11-20T20:30:10.772Z",
        )
        connector._cache.set(
            pk="pointer.test.06dc0fd3c08a2bc6a33f5460da9fea10",
            sk="all",
            value="2022-11-20T20:31:18.382Z",
        )

        # Simulates operations performed by "run()" without calling. Run is not called
        # as this catches all GroveExceptions to prevent unhandled errors in regular
        # operation, which complicates testing.
        connector.save(logs)
        connector.pointer = connector.pointer_next
        connector.save_hashes()

        # Clean-up to simulate a fresh run without having to re-create everything in
        # a subsequent test case.
        connector._cache.delete(
            pk="window_start.test.06dc0fd3c08a2bc6a33f5460da9fea10",
            sk="all",
        )
        connector._cache.delete(
            pk="window_end.test.06dc0fd3c08a2bc6a33f5460da9fea10",
            sk="all",
        )
        connector._cache.delete(
            pk="pointer_next.test.06dc0fd3c08a2bc6a33f5460da9fea10",
            sk="all",
        )
        connector._window_end = ""
        connector._window_start = ""
        connector._pointer_next = ""
        connector._pointer_previous = ""

        # Remove deduplication hashes from memory, to force getting them from cache
        # simulating a real run.
        connector._hashes = {}

        # Simulates a subsequent collection / execution. This is to ensure that logs are
        # collected as expected after the "recovery" operation has complete.
        connector.save(logs)
        connector.pointer = connector.pointer_next
        connector.save_hashes()

        # The pointer should now be the latest log entry - as we should be caught up.
        self.assertEqual(connector.pointer, logs[0].get("time"))



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Zoom API client.

As the Python Zoom client does not currently support Audit API, this client has been
created in the interim.
"""

import base64
import logging
import time
from typing import Any, Dict, Optional

import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://api.zoom.us"


class Client:
    def __init__(
        self,
        identity: Optional[str] = None,
        client_id: Optional[str] = None,
        key: Optional[str] = None,
        retry: Optional[bool] = True,
    ):
        """Setup a new client.

        :param identity: The Zoom account ID
        :param client_id: The Zoom integration client id
        :param key: The Zoom integration client secret
        :param access_token: The allocated Zoom integration client refresh token
        :param retry: Whether to automatically retry if recoverable errors are
            encountered, such as rate-limiting.
        """
        self.identity = identity
        self.key = key
        self.client_id = client_id
        self.retry = retry
        self.logger = logging.getLogger(__name__)

        self.headers = {
            "Content-Type": "application/json",
        }

    def _get(
        self, url: str, params: Optional[Dict[str, Optional[str]]] = None
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: A URL to perform the HTTP GET against.
        :param headers: A dictionary of headers to add to the request.
        :param parameters: An optional set of HTTP parameters to add to the request.

        :return: HTTP Response object containing the headers and body of a response.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.
        """
        while True:
            try:
                response = requests.get(url, headers=self.headers, params=params)
                response.raise_for_status()
                break
            except requests.exceptions.HTTPError as err:
                # Retry on rate-limit, but only if requested.
                if err.response.status_code == 429:
                    self.logger.warning("Rate-limit was exceeded during request")
                    if self.retry:
                        time.sleep(int(err.response.headers.get("Retry-After", "1")))
                        continue
                    else:
                        raise RateLimitException(err) from err

                raise RequestFailedException(err) from err

        return HTTPResponse(headers=response.headers, body=response.json())

    def _post(
        self,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """A POST wrapper to handle retries for the caller.

        :return: the json response.

        :raises RequestFailedException: An HTTP request failed.
        """
        try:
            response = requests.post(
                url,
                headers=headers,
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise RequestFailedException(err) from err

        return response.json()

    def get_access_token(self):
        """Use Basic Auth to get Bearer Token.

        This is required by Zoom to auth the integration and then grant
        the bearer token to access the API.

        To get the access token the accountid, grant_type has to be in the url and
        not in the data.

        :returns: If the request is successful, the bearer token is returned to
            the Client class header.
        """
        grant_type = "account_credentials"
        url = f"https://zoom.us/oauth/token?grant_type={grant_type}&account_id={self.identity}"

        # set basic auth value for authorization header
        basic_auth = str(
            base64.b64encode(bytes(f"{self.client_id}:{self.key}", "utf-8")),
            "utf-8",
        )

        bearer_response = self._post(
            url,
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Basic {basic_auth}",
            },
        )

        access_token = bearer_response.get("access_token")
        self.headers["Authorization"] = f"Bearer {access_token}"

    def get_logs(
        self,
        endpoint: str,
        result_field: str,
        to_date: Optional[str] = None,
        from_date: Optional[str] = None,
        cursor: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of logs from Zoom which match the provided filters.

        :param from_date: The required date of the earliest log entry.
        :param to_date: The required date of the latest log entry.
        :param limit: The maximum number of items to include in a single response.
        :param cursor: The 'next_page_token' returned from a request.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        # The endpoint returns the same total value of results regardless of the limit
        # and offset parameters. The pagination parameters determine the amount of
        # content in the data[] array.

        result = self._get(
            f"{API_BASE_URI}/{endpoint}",
            params={
                "from": from_date,
                "to": to_date,
                "next_page_token": cursor,
            },
        )

        # get the list of logs from the result_field parameters value
        data = result.body.get(result_field, [])

        # keep paging until we meet the total number of results
        cursor = result.body.get("next_page_token", None)
        if cursor == "":
            cursor = None

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=data)

    def get_operationlogs(
        self,
        from_date: Optional[str],
        to_date: Optional[str] = None,
        cursor: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of audit logs from Zoom which match the provided filters.

        :param from_date: The required date of the earliest log entry.
        :param to_date: The required date of the latest log entry.
        :param cursor: The cursor to use when paging.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        return self.get_logs(
            endpoint="v2/report/operationlogs",
            result_field="operation_logs",
            from_date=from_date,
            to_date=to_date,
            cursor=cursor,
        )

    def get_activities(
        self,
        from_date: Optional[str],
        to_date: Optional[str] = None,
        cursor: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of activity logs from Zoom which match the provided filters.

        :param from_date: The required date of the earliest log entry.
        :param to_date: The required date of the latest log entry.
        :param cursor: The cursor to use when paging.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        return self.get_logs(
            endpoint="v2/report/activities",
            result_field="activity_logs",
            from_date=from_date,
            to_date=to_date,
            cursor=cursor,
        )



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Atlassian API client.

As Atlassian does not currently support the Events API, this client has been created in
the interim.
"""

import logging
import time
from typing import Dict, Optional
from datetime import datetime, timezone
import requests

from grove.exceptions import RateLimitException, RequestFailedException
from grove.types import AuditLogEntries, HTTPResponse

API_BASE_URI = "https://api.atlassian.com/admin/v1/orgs/{identity}"
API_DATE_FORMAT = "%Y-%m-%dT%H:%M%z"

class Client:
    def __init__(
        self,
        identity: Optional[str] = None,
        token: Optional[str] = None,
        retry: Optional[bool] = True,
    ):
        """Setup a new client.

        :param identity: The name of the Atlassian organisation.
        :param token: The Atlassian API token.
        :param retry: Automatically retry if recoverable errors are encountered, such as
            rate-limiting.
        """
        self.logger = logging.getLogger(__name__)
        self.retry = retry
        self.headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {token}",
        }

        # We need to push the identity into the URI, so we'll keep track of this.
        self._api_base_uri = API_BASE_URI.format(identity=identity)

    def _get(
        self,
        url: str,
        params: Optional[Dict[str, Optional[str]]] = None,
    ) -> HTTPResponse:
        """A GET wrapper to handle retries for the caller.

        :param url: URL to perform the HTTP GET against.
        :param params: HTTP parameters to add to the request.

        :raises RateLimitException: A rate limit was encountered.
        :raises RequestFailedException: An HTTP request failed.

        :return: HTTP Response object containing the headers and body of a response.
        """
        while True:
            try:
                response = requests.get(url, headers=self.headers, params=params)
                response.raise_for_status()
                break
            except requests.exceptions.RequestException as err:
                if int(getattr(err.response, "status_code", 0)) != 429:
                    raise RequestFailedException(err)

                if not self.retry:
                    raise RateLimitException(err)

                # Retry on rate-limit, but only if requested or if the API does not
                # return a reset at time.
                reset_at = err.response.headers.get("X-Ratelimit-Reset")
                if not reset_at:
                    raise RateLimitException(err)

                self.logger.warning(
                    "Rate-limit was exceeded during collection.",
                    extra={
                        "X-Ratelimit-Reset": reset_at,
                    }
                )

                # Work out how long we need to wait, and if too long, just bail early.
                time_current = int(datetime.now(timezone.utc).strftime("%s"))
                time_reset = int(datetime.strptime(reset_at, API_DATE_FORMAT).strftime("%s")) # noqa: E501
                time_wait = time_reset - time_current

                if time_wait >= 180:
                    raise RateLimitException(err)
                if time_wait <= 0:
                    continue

                time.sleep(time_wait)

        return HTTPResponse(headers=response.headers, body=response.json())

    def get_audit(
        self,
        cursor: Optional[str] = None,
        from_date: Optional[str] = None,
    ) -> AuditLogEntries:
        """Fetches a list of signing attempt logs.

        :param cursor: Cursor to use when fetching results. Supersedes other parameters.
        :param from_date: The earliest date an event represented as a UNIX epoch time.

        :return: AuditLogEntries object containing a pagination cursor, and log entries.
        """
        url = f"{self._api_base_uri}/events"

        # Use the cursor if set, otherwise construct the initial query.
        if cursor is not None:
            self.logger.debug(
                "Collecting next page with provided cursor.", extra={"cursor": cursor}
            )
            result = self._get(url, params={"from": from_date, "cursor": cursor})
        else:
            self.logger.debug(
                "Collecting first page with provided", extra={"from_date": from_date}
            )
            result = self._get(url, params={"from": from_date})

        cursor = result.body.get("meta", {}).get("next", None)

        # Return the cursor and the results to allow the caller to page as required.
        return AuditLogEntries(cursor=cursor, entries=result.body.get("data", []))



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for the GitHub API Client."""

import unittest

from grove.connectors.github.api import Client


class GitHubClientTestCase(unittest.TestCase):
    """Implements tests for the GitHub API Client."""

    def test_client_parse_link_header(self):
        """Ensure Link header parsing works as expected."""
        client = Client(identity="Identity", token="Key")

        # Ensure the next page URL is extracted correctly.
        candidate = (
            '<https://api.github.com/user/repos?page=3&per_page=100>; rel="next", '
            '<https://api.github.com/user/repos?page=50&per_page=100>; rel="last"'
        )
        expected = "https://api.github.com/user/repos?page=3&per_page=100"
        self.assertEqual(client._parse_link_header(candidate), expected)

        # Ensure SSRF mitigation is working correctly.
        candidate = (
            '<http://192.0.2.1:8080/some/vulnerable/service?action=thing>; rel="next", '
            '<https://api.github.com/user/repos?page=50&per_page=100>; rel="last"'
        )
        with self.assertRaises(ValueError):
            client._parse_link_header(candidate)

        # Ensure a ValueError is returned when there are no pages left.
        candidate = (
            '<https://api.github.com/user/repos?page=3&per_page=100>; rel="first", '
            '<https://api.github.com/user/repos?page=50&per_page=100>; rel="previous"'
        )
        with self.assertRaises(ValueError):
            client._parse_link_header(candidate)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Torq activity log connector for Grove."""

from datetime import datetime, timedelta, timezone

from grove.connectors import BaseConnector
from grove.connectors.torq.api import Client
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "torq_activity_logs"
    POINTER_PATH = "event.timestamp"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    def collect(self):
        """Collects all activity logs from the Torq API.

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, a 7 day look-back of data will be collected.
        """
        client = Client(identity=self.identity, key=self.key)
        cursor = None

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to 7-days ago.
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.now(timezone.utc) - timedelta(days=7)).isoformat()

        while True:
            log = client.get_activity_logs(start_time=self.pointer, cursor=cursor)

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor
            if cursor is None:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the Torq Audit Log collector."""

import os
import re
import unittest
from unittest.mock import patch

import responses
from responses import matchers

from grove.connectors.torq.audit_logs import Connector
from grove.models import ConnectorConfig
from tests import mocks


class TorqAuditLogsTestCase(unittest.TestCase):
    """Implements unit tests for the Torq Audit Logs collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                identity="client_id",
                key="client_secret",
                name="test",
                connector="test",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @responses.activate
    def test_collect_pagination(self):
        """Ensure pagination is working as expected."""
        # set bearer token
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/torq/client/001.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
        )
        # first page has a cursor to the next page
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/torq/audit/001.json"), "r"
                ).read(),
                "utf-8",
            ),
            match=[
                matchers.header_matcher(
                    {"Authorization": f"Bearer aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}
                )
            ],
        )

        # The last page returns no cursor.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/torq/audit/002.json"), "r"
                ).read(),
                "utf-8",
            ),
            match=[
                matchers.header_matcher(
                    {"Authorization": f"Bearer aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}
                )
            ],
        )

        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 4)
        # it's reverse chronological so the earlier timestamp should be recorded
        self.assertEqual(self.connector.pointer, "2022-06-27T11:35:10.681687Z")

    @responses.activate
    def test_collect_no_pagination(self):
        """Ensure collection without pagination is working as expected."""
        # set bearer token
        responses.add(
            responses.POST,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/torq/client/001.json"),
                    "r",
                ).read(),
                "utf-8",
            ),
        )

        # add one entry result with no cursor
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/torq/audit/002.json"), "r"
                ).read(),
                "utf-8",
            ),
            match=[
                matchers.header_matcher(
                    {"Authorization": f"Bearer aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}
                )
            ],
        )

        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 1)
        self.assertEqual(self.connector.pointer, "2022-06-07T11:35:11.681687Z")



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove processor to split a log entry into N log entries by the specified JMESPath.

This processor is intended to allow "fanning-out" a single log entry which contains
several related operations into distinct log entries per item. The remainder of the
log entry outside of the split path will not be modified.

It is important to note that the list of elements which is fanned-out will be converted
into a dictionary, rather than a list.

As an example, in following sample:

    {
        "events": [
            {"name": "First", "value": 1},
            {"name": "Second", "value": 2},
        ]
    }

After splitting based on "events", two records would be generated, which contain the
following:

    {
        "events": {"name": "First", "value": 1},
    }

    {
        "events": {"name": "Second", "value": 2},
    }

"""

from typing import Any, Dict, List

import jmespath
from pydantic import Extra

from grove.helpers import parsing
from grove.models import ProcessorConfig
from grove.processors import BaseProcessor


class Handler(BaseProcessor):
    """Split a log entry into N log entries by the specified JMESPath."""

    class Configuration(ProcessorConfig, extra=Extra.forbid):
        """Expresses the configuration and associated validators for the processor."""

        # Source defines the path to split the log entry by. This should be defined as a
        # JMESPath. The field referenced by this path should be a list.
        source: str

    def process(self, entry: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Attempt to extract and map fields from the log entry.

        :param entry: A collected log entry.

        :return: The processed log entry.
        """
        # In this instance we WANT to mutate the copy outside of the processor.
        processed = []
        children = jmespath.search(self.configuration.source, entry)

        # To ensure we don't accidentally try and split a string or dictionary we need
        # to make sure that the type of the found children - if any - is correct.
        if not children or not isinstance(children, list) or len(children) < 1:
            return [entry]

        for child in children:
            processed.append(
                parsing.update_path(
                    parsing.quick_copy(entry),
                    parsing.quote_aware_split(self.configuration.source),
                    child,
                    replace=True,
                )
            )

        return processed



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove processor to extract and map fields using JMESPaths.

This processor is intended to be used to transform raw log entries into a common schema.
This is especially useful for ensuring that all collected log entries from differing
upstream vendors are in a consistent format - whether industry standard, or bespoke.
"""

import json
from typing import Any, Dict, List, Optional

import jmespath
from pydantic import BaseModel, Extra, validator

from grove.helpers import parsing
from grove.models import ProcessorConfig
from grove.processors import BaseProcessor


class Mapping(BaseModel, extra=Extra.forbid):
    """Expresses the configuration fields used to specify path mapping."""

    # Destination specifies where to write extracted or specified values into. This
    # can be a nested path, with subsequent dimensions specified with dots (`.`).
    destination: str

    # Sources defines a list of JMESPaths to map into the destination. If multiple
    # are provided, the sources are processed in order with the first match winning.
    sources: List[str] = []

    # Static allows a static field to be written into the destination, rather than
    # extraction from the source. This field is incompatible with sources.
    static: Optional[str]

    @validator("static")
    def static_or_sources(cls, value, values):
        """Ensures that either sources or static is set, not both."""
        if value and len(values.get("sources")) > 0:
            raise ValueError("Either sources or static should be set, not both.")

        return value


class Handler(BaseProcessor):
    """Extract and map fields using JMESPaths."""

    class Configuration(ProcessorConfig, extra=Extra.forbid):
        """Expresses the configuration and associated validators for the processor."""

        # Remap the original event as a string under the provided path. If not set, any
        # field not mapped will be dropped.
        raw: Optional[str]

        # Defines the field mapping.
        fields: List[Mapping]

    def process(self, entry: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Attempt to extract and map fields from the log entry.

        :param entry: A collected log entry.

        :return: The processed log entry with fields mapped, as a list.
        """
        result: Dict[str, Any] = {}

        # Map the entire log entry under the given path - if configured.
        if self.configuration.raw:
            result = parsing.update_path(
                result,
                parsing.quote_aware_split(self.configuration.raw),
                json.dumps(entry, separators=(",", ":")),
            )

        for field in self.configuration.fields:
            value = field.static
            destination = parsing.quote_aware_split(field.destination)

            # If a static value is defined it should be used over any source fields.
            if not value:
                # Mappings may contain multiple sources to attempt to map. These are
                # evaluated from the first entry to the last, with the first match
                # winning.
                for source in field.sources:
                    value = jmespath.search(source, entry)
                    if value:
                        break

            # Combine the extracted value with the data nested under the same path - or
            # create the path if not present.
            result = parsing.update_path(result, destination, value)

        # Return the newly processed entry. A list is always used, even if only a single
        # element is returned, to allow support for dropping log entries, or splitting a
        # single log entry into multiple.
        return [result]



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements unit tests for the Okta SystemLog API collector."""

import os
import re
import unittest
from unittest.mock import patch

import responses

from grove.connectors.okta.system_log import Connector
from grove.models import ConnectorConfig
from tests import mocks


class OktaAuditTestCase(unittest.TestCase):
    """Implements unit tests for the Okta SystemLog Audit collector."""

    @patch("grove.helpers.plugin.load_handler", mocks.load_handler)
    def setUp(self):
        """Ensure the application is setup for testing."""
        self.dir = os.path.dirname(os.path.abspath(__file__))
        self.connector = Connector(
            config=ConnectorConfig(
                identity="example",
                key="token",
                name="examplecorp",
                connector="test",
            ),
            context={
                "runtime": "test_harness",
                "runtime_id": "NA",
            },
        )

    @responses.activate
    def test_client_rate_limit(self):
        """Ensure ratelimit waiting is working as expected."""
        # Rate limit the first request.
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=429,
            content_type="application/json",
            body=bytes(),
            headers={
                "X-Rate-Limit-Remaining": "0",
                "X-Rate-Limit-Reset": "0",  # Unix-time, so definitely in the past :)
            },
        )

    @responses.activate
    def test_collect_no_pagination(self):
        """Ensure collection without pagination is working as expected."""
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/okta/system_log/001.json"), "r"
                ).read(),
                "utf-8",
            ),
        )
        # Ensure only a single value is returned, and the pointer is properly set.
        self.connector.run()
        self.assertEqual(self.connector._saved["logs"], 1)
        self.assertEqual(self.connector.pointer, "2021-06-24T00:04:08.123Z")

    @responses.activate
    def test_collect_no_results(self):
        """Ensure break occurs when there is no results returned."""
        responses.add(
            responses.GET,
            re.compile(r"https://.*"),
            status=200,
            content_type="application/json",
            body=bytes(
                open(
                    os.path.join(self.dir, "fixtures/okta/system_log/002.json"), "r"
                ).read(),
                "utf-8",
            ),
        )



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove in memory cache handler."""

import logging
from typing import Optional

from grove.caches import BaseCache
from grove.exceptions import DataFormatException, NotFoundException


class Handler(BaseCache):
    """A volatile in-memory backed cache for pointers and other Grove data."""

    def __init__(self):
        self._data = {}
        self.logger = logging.getLogger(__name__)

    def get(self, pk: str, sk: str) -> str:
        """Retrieve a value with the given PK / SK.

        :param pk: Partition key of the value to retrieve.
        :param sk: Sort key of the value to retrieve.

        :raises NotFoundException: No value was found.

        :return: Value from the cache.
        """
        value = self._data.get(pk, {}).get(sk, None)

        if value is None:
            self.logger.info("No value found in cache", extra={"pk": pk, "sk": sk})
            raise NotFoundException("No value found in cache")

        return value

    def set(
        self,
        pk: str,
        sk: str,
        value: str,
        not_set: bool = False,
        constraint: Optional[str] = None,
    ):
        """Stores the value for the given key in a local dict.

        :param pk: Partition key of the value to save.
        :param sk: Sort key of the value to save.
        :param value: Value to save.
        :param not_set: Specifies whether the value must not already be set in the cache
            for the set to be successful.
        :param constraint: An optional condition to use set operation. If provided,
            the currently cached value must match for the delete to be successful.

        :raises ValueError: An incompatible set of parameters were provided.
        :raises DataFormatException: The provided constraint was not satisfied.
        """
        if constraint is not None and not_set:
            raise ValueError("A value cannot both have a constraint AND not be set.")

        # First check if the value is set, and if so whether the caller requires that it
        # NOT already be set.
        current = None

        try:
            current = self._data[pk][sk]
        except KeyError:
            pass

        if current and not_set:
            raise DataFormatException("Value is already set in cache")

        # Next check if the constraint is met.
        if constraint is not None and current != constraint:
            raise DataFormatException("Current value in cache did not match constraint")

        # Finally, set the value.
        if pk not in self._data:
            self._data[pk] = {}

        self._data[pk][sk] = value

    def delete(self, pk: str, sk: str, constraint: Optional[str] = None):
        """Deletes an entry from dict that has the given PK / SK.

        :param pk: Partition key of the value to delete.
        :param sk: Sort key of the value to delete.
        :param constraint: An optional condition to use during the delete. The value
            provided as the condition must match for the delete to be successful.

        :raises DataFormatException: The provided constraint was not satisfied.
        """
        # To enforce constraints, we first need the current value - if any.
        current = None

        try:
            current = self._data[pk][sk]
        except KeyError:
            pass

        # Next check if the constraint is met.
        if constraint is not None and current != constraint:
            raise DataFormatException("Current value in cache did not match constraint")

        if pk in self._data:
            if sk in self._data[pk]:
                del self._data[pk][sk]

            if not self._data[pk]:
                del self._data[pk]



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Grove processor to zip two sets of data into a dictionary of key / value pairs.

This processor is useful for transforming lists of key / value pairs, where the key and
the value of a desired set of data. This processor is useful for flattening key / value
data from sources such as Google Workspace activity logs.

In line with this Google Workspace example, data from Google may appear as follows:

    "parameters": [
        {"name": "owner", "value": "a-user@example.org"},
        {"name": "visibility", "value": "private"}
    ]

Unfortunately, this data can be hard to work with in many SIEMs and search indexes as it
is expressed in the raw log entries. As a result, this processor may be used to instead
'flatten' this key / value data into a dictionary which is keyed by the extracted value
of "name", and uses the value from the "value" field:

    "parameters": {
        "owner": "a-user@example.org",
        "visibility": "private"
    }

Making this data considerably easier to work with during creation of indexes, and
creation of detection content.
"""

from typing import Any, Dict, List

import jmespath
from pydantic import Extra

from grove.helpers import parsing
from grove.models import ProcessorConfig
from grove.processors import BaseProcessor


class Handler(BaseProcessor):
    """Extract and map fields using JMESPaths."""

    class Configuration(ProcessorConfig, extra=Extra.forbid):
        """Expresses the configuration and associated validators for the processor."""

        # Source defines the JMESPath to the data which needs to be zipped.
        source: str

        # Key defines the JMESPath of the data to use as keys in the constructed
        # dictionary. This must be the path relative to the source, not the absolute
        # path.
        key: str

        # Values defines the JMESPaths of the data to use as values in the constructed
        # dictionary. If multiple are provided, the sources are processed in order with
        # the first match winning. This must be the path relative to the source, not the
        # absolute path.
        values: List[str] = []

    def process(self, entry: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Extract and zip configured paths, replacing the source.

        :param entry: A collected log entry.

        :return: The processed log entry with fields zipped.
        """
        result: Dict[str, Any] = {}
        children: List[Any] = []

        # If the source field cannot be found, just pass the record back to the caller
        # as we don't want to drop it. We also want to make sure we can always iterate
        # over the children, so if the value isn't a list, map it into one.
        candidate = jmespath.search(self.configuration.source, entry)
        if candidate is None:
            return [entry]

        if isinstance(candidate, list):
            children = candidate
        else:
            children = [candidate]

        for child in children:
            # No key? Skip.
            key = jmespath.search(self.configuration.key, child)
            if key is None:
                continue

            # No values found? Skip.
            value = None
            for path in self.configuration.values:
                value = jmespath.search(path, child)
                if value is not None:
                    break

            if value is None:
                continue

            # If we have both save it and move on.
            result[key] = value

        # Map the processed data over the top of the original.
        processed = parsing.update_path(
            entry,
            parsing.quote_aware_split(self.configuration.source),
            result,
            replace=True,
        )

        # Return the newly processed entry. A list is always used, even if only a single
        # element is returned, to allow support for dropping log entries, or splitting a
        # single log entry into multiple.
        return [processed]



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Implements tests for the HashiCorp Vault secrets backend."""

import json
import os
import re
import tempfile
import unittest

import responses

from grove.exceptions import AccessException
from grove.secrets.hashicorp_vault import Handler


class SecretsHashiCorpVaultTestCase(unittest.TestCase):
    """Implements tests for the HashiCorp Vault secrets backend."""

    @responses.activate
    def setUp(self):
        self.token = "A_VERY_SECRET_VALUE"

        # Set sensible defaults for the test harness.
        os.environ["GROVE_SECRET_HASHICORP_VAULT_ADDR"] = "http://192.0.2.1:8200"
        os.environ["GROVE_SECRET_HASHICORP_VAULT_TOKEN"] = self.token

        # Mock the pre-flight to always succeed (used to validate credentials).
        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=200,
            content_type="application/json",
            body=bytes(),
        )
        self.secrets = Handler()

    @responses.activate
    def test_token_from_file(self):
        """Ensures Vault token can be read from file."""
        expected = "THIS_IS_A_SECRET_FROM_FILE"

        # Mock the pre-flight to always succeed (used to validate credentials).
        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=200,
            content_type="application/json",
            body=bytes(),
        )

        with tempfile.NamedTemporaryFile("w") as fout:
            fout.write(expected)
            fout.write("\n")
            fout.flush()

            # Setup a new Vault handler using this file.
            os.environ["GROVE_SECRET_HASHICORP_VAULT_TOKEN_FILE"] = fout.name

            secrets = Handler()

        # Check that the token on the configured / setup secrets handler matches the
        # value written to the temporary file.
        self.assertEqual(secrets.config.token, expected)

    @responses.activate
    def test_client_setup(self):
        """Ensure the client validates credentials on setup."""
        # Pre-flight should fail if Vault is not reachable (no mock, and TEST-NET-1 URL)
        with self.assertRaises(AccessException):
            _ = Handler()

        # Mock a successful pre-flight - used to validate credentials
        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=200,
            content_type="application/json",
            body=bytes(),
        )
        secrets = Handler()

        # Ensure that the token on the handler matches the expected value.
        secrets.config.token = self.token

        # Mock an unsuccessful pre-flight.
        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=401,
            content_type="application/json",
            body=bytes(),
        )
        with self.assertRaises(AccessException):
            _ = Handler()

    @responses.activate
    def test_get_field_and_path(self):
        """Ensure that field and path extraction is working as expected."""
        # Ensure fields are extracted correctly.
        candidate = "secret/data/example/demo?field=password"
        field, path = self.secrets.get_field_and_path(candidate)

        self.assertEqual(field, "password")
        self.assertEqual(path, "secret/data/example/demo")

        # Ensure omitted fields raise an exception.
        candidate = "secret/data/example/demo"
        with self.assertRaises(ValueError):
            self.secrets.get_field_and_path(candidate)

    @responses.activate
    def test_get_field_v1(self):
        """Ensure that a KVv1 secret can be fetched."""
        candidate = "secret/example/demo?field=password"
        expected = "SUPER_SECRET_VALUE"

        # Positive case - field and secret exists.
        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=200,
            content_type="application/json",
            body=bytes(json.dumps({"data": {"password": expected}}), "utf-8"),
        )

        secret = self.secrets.get(candidate)
        self.assertEqual(secret, expected)

        # Negative case - incorrect / missing field should raise an error.
        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=200,
            content_type="application/json",
            body=bytes(json.dumps({"data": {"unknown": expected}}), "utf-8"),
        )

        with self.assertRaises(AccessException):
            _ = self.secrets.get(candidate)

    @responses.activate
    def test_get_error(self):
        """Ensure authentication errors are handled."""
        candidate = "secret/data/example/demo?field=token"

        # Negative case - incorrect / missing field should raise an error.
        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=401,
            content_type="application/json",
            body=bytes(),
        )

        with self.assertRaises(AccessException):
            _ = self.secrets.get(candidate)

    @responses.activate
    def test_get_field_v2(self):
        """Ensure that a KVv2 secret can be fetched."""
        candidate = "secret/data/example/demo?field=token"
        expected = "SUPER_SECRET_VALUE_TWO"

        # Positive case - field and secret exists.
        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=200,
            content_type="application/json",
            body=bytes(json.dumps({"data": {"data": {"token": expected}}}), "utf-8"),
        )

        secret = self.secrets.get(candidate)
        self.assertEqual(secret, expected)

        # Negative case - incorrect / missing field should raise an error.
        responses.add(
            responses.GET,
            re.compile(r"http://.*"),
            status=200,
            content_type="application/json",
            body=bytes(json.dumps({"data": {"data": {"unknown": expected}}}), "utf-8"),
        )

        with self.assertRaises(AccessException):
            _ = self.secrets.get(candidate)



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""PagerDuty Audit connector for Grove."""

from datetime import datetime, timedelta

from grove.connectors import BaseConnector
from grove.connectors.pagerduty.api import Client
from grove.constants import REVERSE_CHRONOLOGICAL
from grove.exceptions import NotFoundException


class Connector(BaseConnector):
    NAME = "pagerduty_audit_records"
    POINTER_PATH = "execution_time"
    LOG_ORDER = REVERSE_CHRONOLOGICAL

    def collect(self):
        """Collects PagerDuty audit records from the PagerDuty API.

        https://developer.pagerduty.com/api-reference/

        This will first check whether there are any pointers cached to indicate previous
        collections. If not, the last week of data will be collected.
        """
        client = Client(token=self.key)
        cursor = None

        # If no pointer is stored then a previous run hasn't been performed, so set the
        # pointer to a week ago. In the case of the PagerDuty audit API the pointer is
        # the value of the "execution_time" field from the latest record retrieved from
        # the API - which is in ISO8601 Format.
        try:
            _ = self.pointer
        except NotFoundException:
            self.pointer = (datetime.utcnow() - timedelta(days=7)).isoformat()

        # Get log data from the upstream API, paging as required.
        while True:
            log = client.get_records(since=self.pointer, cursor=cursor)

            # Save this batch of log entries.
            self.save(log.entries)

            # Check if we need to continue paging.
            cursor = log.cursor
            if cursor is None:
                break



# Copyright (c) HashiCorp, Inc.
# SPDX-License-Identifier: MPL-2.0

"""Provides collected Grove log output to supported destinations."""

import abc
import gzip
import json
import logging
from typing import Any, Dict, List, Optional

from pydantic import BaseSettings, Extra, ValidationError

from grove.constants import GROVE_METADATA_KEY
from grove.exceptions import ConfigurationException, DataFormatException
from grove.helpers import parsing


class BaseOutput(abc.ABC):
    """The basis for all Grove output handlers."""

    class Configuration(BaseSettings, extra=Extra.allow):
        """Defines the configuration directives required by all output handlers."""

        pass

    def __init__(self):
        """Implements core logic which applies to all handlers.

        This includes configuration of logging, and parsing of configuration.
        """
        self.logger = logging.getLogger(__name__)

        # Wrap validation errors to keep them in the Grove exception hierarchy.
        try:
            self.config = self.Configuration()
        except ValidationError as err:
            raise ConfigurationException(parsing.validation_error(err))

    def setup(self):
        """Implements logic to setup any required clients, sockets, or connections.

        If not required for the given output handler, this may be a no-op.
        """
        pass

    @abc.abstractmethod
    def submit(
        self,
        data: bytes,
        connector: str,
        identity: str,
        operation: str,
        part: int = 0,
        suffix: Optional[str] = None,
        descriptor: Optional[str] = None,
    ):
        """Implements logic require to write collected log data to the given backend.

        :param data: Log data to write.
        :param connector: Name of the connector which retrieved the data.
        :param identity: Identity the collected data was collect for.
        :param operation: Operation the collected logs are associated with.
        :param part: Number indicating which part of the same log stream this file
            contains data for. This is used to indicate that the logs are from the same
            collection, but have been broken into smaller files for downstream
            processing.
        :param suffix: An optional suffix to allow propagation of file type information
            or other relevant features.
        :param descriptor: An optional and arbitrary descriptor associated with the
            log data. This may be used by handlers for construction / specification of
            file paths, URLs, or database tables.
        """
        pass

    def serialize(self, data: List[Any], metadata: Dict[str, Any] = {}) -> bytes:
        """Implements serialization of log entries to a gzipped NDJSON.

        :param data: A list of log entries to serialize to JSON.
        :param metadata: Metadata to append to each log entry before serialization. If
            not specified no metadata will be added.

        :return: Log data serialized as gzipped NDJSON (as bytes).

        :raises DataFormatException: Cannot serialize the input to JSON.
        """
        candidate = []

        # Append the Grove metadata to each log entry, and serialize to JSON. Adding
        # This is expensive but we can't just json.dumps into gzip.compress as that
        # will not yield NDJSON.
        for entry in data:
            # Skip empty log entries.
            if entry is None:
                continue

            if metadata:
                entry[GROVE_METADATA_KEY] = {
                    **metadata,
                    **entry.get(GROVE_METADATA_KEY, {}),
                }

            # We don't want to silently drop and lose single records, so drop the entire
            # batch if there is bad data (which will trigger a retry next run).
            try:
                candidate.append(json.dumps(entry, separators=(",", ":")))
            except TypeError as err:
                raise DataFormatException(f"Unable to serialize to JSON: {err}")

        return gzip.compress(bytes("\r\n".join(candidate), "utf-8"))
